local _, ZbGm = ...

-- Convert date to string
local function _StringFromDate(dateValue)
    if dateValue ~= nil and type(dateValue) == "number" and dateValue ~= 0 then
	    return date("%m-%d-%y", dateValue, dateValue, dateValue)
    else
        return "   NA"
    end
end


-------------------------------------------------------------------------------------
-- INTERNAL ROSTER MEMORY OBJECT
-------------------------------------------------------------------------------------
ZbGm.ZRoster = {
    -- Data
    players = {},  			-- structure containing core player data and list of characters array.
    playerIndex = {},		-- filtered index into the players array for what is showing on main display.
	
	IndexMaxLen = 29,
	
	-- These functions are used by Table.Sort to sort by various columns.
	sortByFunctions = { 
		["Name"] = function(a,b) return a < b end,
		["NameDesc"] = function(a,b) return a > b end,
		["Log"] = function(a,b) return ZbGm.ZRoster:GetLastLogin(ZbGm.ZRoster.players[a]) <  ZbGm.ZRoster:GetLastLogin(ZbGm.ZRoster.players[b]) 
			end,
		["LogDesc"] = function(a,b) return ZbGm.ZRoster:GetLastLogin(ZbGm.ZRoster.players[b]) <  ZbGm.ZRoster:GetLastLogin(ZbGm.ZRoster.players[a])  end,
		["Join"] = function(a,b) return ZbGm.ZRoster:GetJoinDate(ZbGm.ZRoster.players[a]) < ZbGm.ZRoster:GetJoinDate(ZbGm.ZRoster.players[b])  end,
		["JoinDesc"] = function(a,b) return ZbGm.ZRoster:GetJoinDate(ZbGm.ZRoster.players[a]) > ZbGm.ZRoster:GetJoinDate(ZbGm.ZRoster.players[b])  end,
		["Rank"] = function(a,b) 
			if ZbGm.ZRoster.players[a].rank == ZbGm.ZRoster.players[b].rank then
				-- sort by name if equal rank.
				return ZbGm.ZRoster.sortByFunctions["Name"](a,b);
			else
				return ZbGm.ZRoster.players[a].rank < ZbGm.ZRoster.players[b].rank;
			end
		end,
		["RankDesc"] = function(a,b) 
			if ZbGm.ZRoster.players[a].rank == ZbGm.ZRoster.players[b].rank then
				-- sort by name if equal rank.
				return ZbGm.ZRoster.sortByFunctions["Name"](a,b);
			else
				return ZbGm.ZRoster.players[a].rank > ZbGm.ZRoster.players[b].rank;
			end
		end,
		["Alts"] = function(a,b) return ZbGm.ZRoster:GetNumAlts(ZbGm.ZRoster.players[a]) < ZbGm.ZRoster:GetNumAlts(ZbGm.ZRoster.players[b])  end,
		["AltsDesc"] = function(a,b) return ZbGm.ZRoster:GetNumAlts(ZbGm.ZRoster.players[a]) > ZbGm.ZRoster:GetNumAlts(ZbGm.ZRoster.players[b]) end,
		["Realm"] = function(a,b) return ZbGm.ZRoster.players[a].server < ZbGm.ZRoster.players[b].server end,
		["RealmDesc"] = function(a,b) return ZbGm.ZRoster.players[a].server > ZbGm.ZRoster.players[b].server end,		
	},
	
	sortIndexBy = "Name",   -- "Name", "Rank", "Joined", "Log", "Alts"
	
	-- Has table of all characters
	--    Character (name, server, rank, class, lastlog, joindate)
	--       parent
	--       child ->nextChild
	
	AddCharacter = function(self, parentIndex, name, server, rank, class, note, officernote, login, join) 
		local full = string.format("%s-%s", name, server);
		local myIndex = full:sub(1, self.IndexMaxLen);
		if parentIndex then 
			parentIndex = parentIndex:sub(1, self.IndexMaxLen) 
		end
		
		--print ("add index = " .. myIndex);
		
		if self.players[myIndex] == nil then
			-- Add Root Record
			self.players[myIndex] = {name=name, server=server, full=full, rank=rank, class=class, note=note, onote=officernote, lastlogin=login, joindate=join, parentIndex=parentIndex, parentNode=nil, childNode=nil, nextChild = nil};
			
			--print(self.players[myIndex].parentIndex)
		else
			-- All other fields don't change
			self.players[myIndex].lastlogin = login;
			self.players[myIndex].joindate = join;
			self.players[myIndex].note = note;
			self.players[myIndex].officernote = officernote;
			self.players[myIndex].parentIndex = parentIndex;				
			self.players[myIndex].childNode = nil;  -- reset lineage
			self.players[myIndex].parentNode = nil; -- reset lineage
			self.players[myIndex].nextChild = nil;
		end
    end,
	
	-- Builds an array of "indexes" to display in the main display.
	-- searchfor allows searching main or alt names.
	BuildFilteredIndex = function(self, searchfor, mains) 
		searchfor = searchfor:lower();
		local newMainList = {};
		
		--mains = true
		
		for idx in pairs(self.players) do
		
			if string.len(searchfor) == 0 then
				if not mains or (mains and self.players[idx].parentNode == nil) then
					table.insert(newMainList, idx)	
				end
			else		
				if string.find(idx:lower(), searchfor) then
					table.insert(newMainList, idx)			
				end
			end
		end
		
		ZbGm.ZRoster.playerIndex = newMainList
		
		--print ZbGm.ZRoster.playerIndex[1]
		self.SortIndex(self)
	end,	
	
	HasUnassociated = function(self) 
		for idx in pairs(self.players) do
			if type(self.players[idx])=="table" then
				if not self.players[idx].parentNode and self.players[idx].joindate == 0 then
					return true;
				end
			end
		end
		return false;
	end,
	
	BuildUnassociatedList = function(self) 
		local unassoc = nil;
		for idx in pairs(self.players) do
			local itm = self.players[idx];
			if type(itm)=="table" then			
				if not itm.parentNode and itm.joindate == 0 and itm.full:sub(1,self.IndexMaxLen) == idx then
					if not unassoc then unassoc = {} end
					table.insert(unassoc, idx)
				end
			end
		end
		return unassoc;
	end,	
	
	ExportCSV = function(self) 
		local exportData = {};
		for idx in pairs(self.players) do
			local itm = self.players[idx];
			
			if itm.parentNode == nil then
				exportData[#exportData+1] = string.format("\"%s\",\"%s\",\"%s\",\"%s\",\"%s\"", itm.full, itm.server, itm.rank, _StringFromDate(itm.joindate), _StringFromDate(itm.lastlogin));
			end
		end		
		return table.concat(exportData, "\n");
	end,
	
	-- Returns a player's main character record.  If the player's main character
	-- is missing, returns the first available character of the player.
	GetFilteredCharacter = function(self, index)
		return ZbGm.ZRoster.players[ZbGm.ZRoster.playerIndex[index]];
	end,
	
	GetCharacter = function(self, charIndex)
		return ZbGm.ZRoster.players[charIndex:sub(1, self.IndexMaxLen)];
	end,
	
	-- Fills up an array with character data.
	GetCharacterIndices = function(self, character, charTable)
	
		if character.parentNode then
			character = character.parentNode;
		end
		
		if character and charTable then
			local i = 1;
			local ptr = character.childNode;

			charTable[i] = character.full;
			
			while ptr do
				i = i + 1;
				charTable[i] = ptr.full;
				--print (i .. " " .. ptr.full);
				ptr = ptr.nextChild;
			end
			return i;
		end
		
		return 0;
	end,
	
	GetNumAlts = function(self, node)
		local parentNode = node;
		if node.parentNode then
			parentNode = node.parentNode;
		end
		
		-- see if already counted;
		if parentNode.childCount then
			return parentNode.childCount;
		end
		
		
		-- unknown figure it out.
		local count = 1;
		local ptr = nil;
		if node.parentNode then
			ptr = node.parentNode.childNode;
		else
			ptr = node.childNode;
		end
		
		while ptr do
			count = count + 1;
			ptr = ptr.nextChild;
		end
		
		parentNode.childCount = count;
		return count			
	end,
	
	GetJoinDate = function(self, node) 
		if node.parentNode then
			return node.parentNode.joindate;			
		else
			return node.joindate;
		end		
	end,
	
	GetLastLogin = function(self, node)
		if node then 
			if node.parentNode then
				return node.parentNode.lastlogin;			
			else
				return node.lastlogin;
			end			
		else
			return 0;
		end
	end,
	
	GetFilteredIndexCount = function(self)
        return #(self.playerIndex)
	end,
	
	ReconstructLists = function(self, savedDatesDB) 
		local fixNotesList = nil;
		local nilOut = nil;
	
		-- Pass 1 - link parent to children.
		-- Pass 2 - clean-up, deal with missing data, call back for updates.
	
		for idx in pairs(self.players) do
			-- Must deal with missing parent.
			-- Must delete missing node			
			
			-- Get a node.
			local node = self.players[idx];

			-- If it has a parent (it isn't a parent).
			if node.parentIndex then
			
				-- Find the node specified as parent.
				local parentNode = self.players[node.parentIndex];
							
				-- If parent node is missing, link to this node as the parent.
				if parentNode == nil then
				
					-- Create a reference by the old parent name to this node as the parent... 
					-- i.e. a false miss-match.  ["jack"] = "harry".
					-- This allows all future missing-parent nodes to find "jack" and creates "harry" as the new parent.
					-- Problem is, the child will still think it's parent is "jack".  This will have to be resolved.
					self.players[node.parentIndex] = node;
					
					-- This is essentially making first found the new main.  So, set this as main.
					if savedDatesDB[node.parentIndex] then
						node.joindate = savedDatesDB[node.parentIndex].join;
						print("zbGuildManager: Recovered missing main join date for " .. node.parentIndex .. " as " .. node.joindate);
						node.parentIndex = nil;
						
						if not fixNotesList then fixNotesList = {} end
						table.insert(fixNotesList, idx);
					end
					
					-- Add a new name to the nil them out list.
					if not nilOut then nilOut = {} end;
					table.insert(nilOut, node.parentIndex);
					--print("nil out - " .. node.parentIndex);
					--print(savedDatesDB[node.parentIndex].join);
				else		
					self:InsertChildNode(node, parentNode);
				end				
			end
			
			if node.parentNode then
				node.parentNode.lastlogin = math.max(node.parentNode.lastlogin, node.lastlogin);
			end
			
		end
		
		-- Blank out the orphaned parents.
		if nilOut then
			for i=1,#nilOut do
				self.players[nilOut[i]] = nil;
			end
		end
		
		-- Count members in time range.
		ZRoster.ActiveCount = 0;
		ZRoster.SemiActiveCount = 0;
		ZRoster.InactiveCount = 0;
		ZRoster.AbsentCount = 0;
		
		for idx in pairs(self.players) do
			if self.players[idx].paretNode == nil then
				local deltaLog = (time() - self.players[idx].lastlogin) / (86400*7)
				--print (deltaLog)
				if deltaLog <= 2 then   		-- Green
					ZRoster.ActiveCount = ZRoster.ActiveCount + 1;	
				elseif deltaLog <= 4 then  	-- Yellow
					ZRoster.SemiActiveCount = ZRoster.SemiActiveCount + 1;
				elseif deltaLog < 19 then  	-- Orange
					ZRoster.InactiveCount = ZRoster.InactiveCount + 1;
				else						-- Red
					ZRoster.AbsentCount = ZRoster.AbsentCount + 1;
				end
			end
		end
		
		--print ("Active = " .. ZRoster.ActiveCount);
		
		return fixNotesList;
	end,
	
	InsertChildNode = function(self, node, parentNode)
		if node.full == parentNode.full then
			return;
		end 	
	
		-- Link child to it's parent.
		node.parentNode = parentNode;
		node.childNode = nil;
		node.nextChild = nil;
		
		-- First child
		if parentNode.childNode == nil then
			parentNode.childNode = node
		else							
			-- Initialize location in parent's children list.
			local curNode = parentNode.childNode;
		
			-- Insert at head of list.
			if node.full < parentNode.childNode.full then
				node.nextChild = parentNode.childNode;
				parentNode.childNode = node;
			else	
			
				while (curNode.nextChild) and node.full > curNode.nextChild.full  do			
					curNode = curNode.nextChild
				end	
				
				node.nextChild = curNode.nextChild
				curNode.nextChild = node
			end
		end
	end, 
	
	RemoveChildNode = function(self, node) 
		local parentNode = node.parentNode;
		
		if parentNode then
			print("remove child node");
			if node.parentNode.childNode == node then
				-- removed!
				node.parentNode.childNode = node.nextChild;
				parentNode.childCount = parentNode.childCount - 1;
			else
				-- Find node before this node.
				local ptr = node.parentNode.childNode;
					
				while ptr do
					if ptr.nextChild == node then
						-- removed!						
						ptr.nextChild = node.nextChild;
						parentNode.childCount = parentNode.childCount - 1;						
						break;
					end
					ptr = ptr.nextChild;
				end					
			end
		end
	end,
	
	RemoveCharacter = function(self, indexName)
		-- Locate the character
		local toon = self.players[indexName];
		local parentNode = nil;
		
		if toon then
			parentNode = toon.parentNode;			
			
			-- If child character, just fix linked list.
			if parentNode then
			
				-- Toon is first child.
				if toon.parentNode.childNode == toon then
					-- removed!
					toon.parentNode.childNode = toon.nextChild;
				else
					-- Find node before this node.
					local ptr = toon.parentNode.childNode;
					
					while ptr do
						if ptr.nextChild == toon then
							-- removed!
							ptr.nextChild = toon.nextChild;
							break;
						end
						ptr = ptr.nextChild;
					end					
				end
			else
				-- IT IS THE PARENT.
				--print ("MAIN REMOVE")
				parentNode = toon.childNode;
				
				if parentNode then
					-- Make child make it's next a child.
					parentNode.childNode = parentNode.nextChild;
					parentNode.nextChild = nil; -- no longer child list.
					parentNode.parentNode = nil; -- no longer a child.
					
					parentNode.joindate = toon.joindate;  -- save the join date.					
					
					local ptr = parentNode.childNode;
					while ptr do
						ptr.parentNode = parentNode;
						ptr = ptr.nextChild;
					end
					
					--print ("new parent " .. parentNode.full)
					--print ("next child " .. parentNode.childNode.full)
				end
			end			
		end
		
		-- Delete the reference to the node.
		self.players[indexName] = nil;
		
		return parentNode;
	end,
	
	SetRank = function(self, index, rank) 
		self.players[index].rank = rank;
	end, 
	
	-- Changes structure where this node is a child of the parent.
	SetMain = function (self, childIndex, parentIndex)
		local childNode = self.players[childIndex:sub(1, self.IndexMaxLen)];
		local parentNode = self.players[parentIndex:sub(1, self.IndexMaxLen)];
		
		if childNode and parentNode then	
			self:InsertChildNode(childNode, parentNode);
			childNode.parentIndex = parentIndex:sub(1, self.IndexMaxLen);
		end
		
		if parentNode.childCount then
			parentNode.childCount = parentNode.childCount + 1;
		end
	end,
	
	SortIndex = function(self)
		table.sort(self.playerIndex, self.sortByFunctions[self.sortIndexBy])
	end,	
	
	SwapMain = function(self, indexOldMain, indexNewMain)
		local oldMain = self.players[indexOldMain:sub(1,self.IndexMaxLen)];
		local newMain = self.players[indexNewMain:sub(1,self.IndexMaxLen)];
		
		-- ensure main node has join date.
		local joindate = oldMain.joindate;
		newMain.joindate = joindate;
		oldMain.joindate = 0;
		
		-- Fix main pointers.
		newMain.parentNode = nil;
		
		local ptr = oldMain.childNode;
		while ptr do
			local saveNext = ptr.nextChild;
			if ptr ~= newMain then 	
				self:InsertChildNode(ptr, newMain); 
			end
			ptr = saveNext;
		end
		
		self:InsertChildNode(oldMain, newMain);
		
		newMain.nextChild = nil;
		
    end,	
	
	UpdateHistorySave = function(self, memberDB) 
		if not memberDB then memberDB = {} end
		
		local now = time();
		for idx in pairs(self.players) do
			if self.players[idx].parentNode == nil then
				if memberDB[idx] then
					memberDB[idx].join=self.players[idx].joindate;
					memberDB[idx].update = now;
				else
					memberDB[idx] = {join=self.players[idx].joindate, update=now};					
				end
			end
		end
		
		-- remove data over a year.
		for idx in pairs(memberDB) do
			if (now - memberDB[idx].update) > 86400*30*2 then
				memberDB[idx] = nil;
			end
		end
	end,
	
	DumpPlayer = function(self, mainIndex)
		-- Get a node.
		local node = self.players[mainIndex];
		
		print ("main = " .. node.name);
		local child = node.childNode;
		
		while child do
			print ("   " .. child.name)
			child = child.nextChild;
		end
	end,
}

ZRoster = ZbGm.ZRoster


